#!/usr/bin/perl
#
# IMSbullToIDC.pl - loads IMS1.0 bulletin into IDC Oracle database schema
#
# Istvan Bondar, 2018
# ibondar2014@gmail.com
#
# Loads an IMS1.0/ISF1.0 or ISF2.0 bulletin file into IDC Oracle DB schema.
# Tables populated:
#     origin
#     origerr
#     assoc
#     arrival
#     amplitude
#     netmag
#     stamag
#     event (if exists)
# Assumes that the last origin in the origin block is the preferred origin
# and associates the arrivals to this preferred origin
#
$|=1;

#
#  required:
#
use strict;
use DBI;
use POSIX;
use Time::Local;

my (%pars, $infile, $machine, $usr, $passwd, $n, $dbh);
#
# check arguments, provide usage info and terminate if there are
# incorrect number of arguments
#
if (@ARGV < 4) {
    print "Usage: LoadIMSbullToIDCOracle.pl inBull machine user password\n";
    exit -1;
}
($infile, $machine, $usr, $passwd) = @ARGV;
#
# DB tables
#
$pars{event_table} = "event";
$pars{origin_table} = "origin";
$pars{origerr_table} = "origerr";
$pars{netmag_table} = "netmag";
$pars{stamag_table} = "stamag";
$pars{assoc_table} = "assoc";
$pars{arrival_table} = "arrival";
$pars{amplitude_table} = "amplitude";
$pars{deltim} = 1;
$pars{delaz} = 15;
$pars{delslo} = 3.5;

#
# Connect to the IDC database
#
$dbh = DBI->connect("dbi:Oracle:$machine", $usr, $passwd) or
            die "Not able to connect to database $usr\@$machine\n";
#
# Parse bulletin
#
$n = ParseBulletin($infile, $dbh);
print "inserted $n events\n";
#
# Disconnect from the IDC database
#
$dbh->disconnect;
print "done\n";

#
#
# Parse bulletin in IMS1.0, ISF1.0 or ISF2.0 format and load into the IDC DB
#
#
sub ParseBulletin {
    my ($infile, $dbh) = @_;
    my ($nev, $eventid, $line, $n);
    my (@originblock, @magblock, @phaseblock);
    open (IN, $infile) || die "Cannot open $infile: $!\n";
#
#   parse IMS bulletin
#
    $nev = 0;
    while (<IN>) {
        s/\r\n$/\n/;      # replace CRLF with LF
        if (/^Event/) {
            $nev++;
            @originblock = @magblock = @phaseblock = ();
#
#           Event block
#
            (undef, $eventid, undef) = split;
            print "found event $eventid\n";
            while (<IN>) {
                s/\r\n$/\n/;      # replace CRLF with LF
                if (/Latitude Longitude/) {
#
#                   Origin block
#
                    while ($line = <IN>) {
                        $line =~ s/\r\n$/\n/;      # replace CRLF with LF
                        next if ($line =~ /^ \(/ or $line =~ /^$/);
                        last if ($line =~ /^Magnitude/ or
                                 $line =~ /^Sta/ or
                                 $line =~ /^STOP/ or
                                 $line =~ /^Year/);
                        push @originblock, $line;
                    }
                    if (/^Magnitude/ or $line =~ /^Magnitude/) {
#
#                       Magnitude block
#
                        while ($line = <IN>) {
                            $line =~ s/\r\n$/\n/;      # replace CRLF with LF
                            next if ($line =~ /^ \(/ or $line =~ /^$/);
                            last if ($line =~ /^Sta/ or $line =~ /^STOP/);
                            push @magblock, $line;
                        }
                    }
                    if (/^Sta/ or $line =~ /^Sta/) {
#
#                       Arrival block
#                           empty line ends arrival block
#
                        while ($line = <IN>) {
                            $line =~ s/\r\n$/\n/;      # replace CRLF with LF
                            next if ($line =~ /^ \(/);
                            last if ($line =~ /^$/);
                            push @phaseblock, $line;
                        }
#
#                       insert event into DB
#
                        print "insert event $nev\n";
                        insert_event($dbh, \@originblock, \@magblock, \@phaseblock);
                    }
                    if (/^STOP/ or $line =~ /^STOP/) {
                        last;
                    }
                    last;
                }
            }
        }
    }
    close IN;
    return $nev;
}

#
#
# insert event into database
#
#
sub insert_event {
    my ($dbh, $originblock, $magblock, $phaseblock) = @_;
    my ($evid, $orid, $magid, $ampid, $arid, $smagid, $mbid, $mlid, $msid);
    my ($htime, $etime, $atime, $jdate, $oday);
    my ($s, $origid, $magorig, $mtyp, $sta, $phase, $isdef, $isamp, $ismag);
    my ($amp, $per, $nass, $auth, $pauth, $delta, $x, $y, $t, $i, $n);
    my ($qevent, $qorig, $qoerr, $qnetmag, $qstamag, $qassoc, $qarriv, $qamplit);
    my ($vorig, $voerr, $vnetmag, $vstamag, $vassoc, $varriv, $vamplit);
    $n = @$originblock;
    if ($n < 1) {
        print "No origin record has been found!\n";
        return;
    }
#
#   get evid from lastid table
#
    $evid = sql_lastid($dbh, "lastid", "evid") + 1;
#
#   parse and insert origins, origerrs and netmags
#
    for ($i = 0; $i < $n; $i++) {
        $s = $originblock->[$i];
        $orid = sql_lastid($dbh, "lastid", "orid") + 1;
#
#       origin record
#
        $qorig = "insert into $pars{origin_table} (evid, orid, ";
        $vorig = "values ($evid, $orid, ";
        $auth = substr($s, 118, 9); $auth =~ s/\ //g;
        if ($auth ne "") { $qorig .= "auth, "; $vorig .= "\'$auth\', "; }
        $htime = substr($s, 0, 22); $htime =~ s/\//\-/g;
        if ($htime eq "") {
            print "Warning: no origin time! Skipping origin.\n$s\n";
            next;
        }
        $etime = human_to_epoch($htime);
        ($htime, $jdate, undef) = epoch_to_human($etime);
        $qorig .= "time, jdate, "; $vorig .= "$etime, $jdate, ";
        $x = substr($s, 36, 8);  $x =~ s/\ //g;
        if ($x ne "") { $qorig .= "lat, "; $vorig .= "$x, "; }
        $x = substr($s, 45, 9);  $x =~ s/\ //g;
        if ($x ne "") { $qorig .= "lon, "; $vorig .= "$x, "; }
        $x = substr($s, 71, 5);  $x =~ s/\ //g;
        if ($x ne "") { $qorig .= "depth, "; $vorig .= "$x, "; }
        $t = substr($s, 76, 1);
        $t = "f" if ($t eq " ");
        $qorig .= "dtype, "; $vorig .= "\'$t\', ";
        $x = substr($s, 83, 4); $x =~ s/\ //g;
        if ($x ne "") { $qorig .= "ndef, "; $vorig .= "$x, "; }
        $t = substr($s, 115, 2); $t =~ s/\ //g;
        if ($t ne "") { $qorig .= "etype, "; $vorig .= "\'$t\', "; }
        $origid = substr($s, 128, 11);       $origid =~ s/\ //g;
#
#       origerr record
#
        $qoerr = "insert into $pars{origerr_table} (orid, ";
        $voerr = "values ($orid, ";
        $x = substr($s, 24, 5); $x =~ s/\ //g;
        if ($x ne "") { $qoerr .= "stime, "; $voerr .= "$x, "; }
        $x = substr($s, 30, 5); $x =~ s/\ //g;
        if ($x ne "") { $qoerr .= "sdobs, "; $voerr .= "$x, "; }
        $x = substr($s, 55, 5); $x =~ s/\ //g;
        if ($x ne "") { $qoerr .= "smajax, "; $voerr .= "$x, "; }
        $x = substr($s, 61, 5); $x =~ s/\ //g;
        if ($x ne "") { $qoerr .= "sminax, "; $voerr .= "$x, "; }
        $x = substr($s, 67, 3); $x =~ s/\ //g;
        if ($x ne "") { $qoerr .= "strike, "; $voerr .= "$x, "; }
        $x = substr($s, 78, 4); $x =~ s/\ //g;
        if ($x ne "") { $qoerr .= "sdepth, "; $voerr .= "$x, "; }
        $qoerr .= "lddate"; $voerr .= "sysdate";
        $qoerr .= ") $voerr)";
        sql_do($dbh, $qoerr) || die "Failed query: $qoerr\n";
#
#       netmag record
#
        for $s (@$magblock) {
            $magorig = substr($s, 30, 11); $magorig =~ s/\ //g;
            next if ($magorig ne $origid);
            $magid = sql_lastid($dbh, "lastid", "magid") + 1;
            $qnetmag = "insert into $pars{netmag_table} (evid, orid, magid, ";
            $vnetmag = "values ($evid, $orid, $magid, ";
            $mtyp = substr($s, 0, 6);  $mtyp =~ s/\ //g;
            if ($mtyp ne "") { $qnetmag .= "magtype, "; $vnetmag .= "\'$mtyp\', "; }
            $x = substr($s, 6, 4); $x =~ s/\ //g;
            if ($x ne "") { $qnetmag .= "magnitude, "; $vnetmag .= "$x, "; }
            $smagid = $magid;
            if ($mtyp eq "mb" or $mtyp eq "mb_ave" or $mtyp eq "mbmle") {
                $qorig .= "mb, mbid, "; $vorig .= "$x, $magid, ";
                $mbid = $magid;
            }
            if ($mtyp eq "Ms" or $mtyp eq "MS" or $mtyp eq "ms_ave" or $mtyp eq "msmle") {
                $qorig .= "ms, msid, "; $vorig .= "$x, $magid, ";
                $msid = $magid;
            }
            if ($mtyp eq "ML") {
                $qorig .= "ml, mlid, "; $vorig .= "$x, $magid, ";
                $mlid = $magid;
            }
            $x = substr($s, 11, 3); $x =~ s/\ //g;
            if ($x ne "") { $qnetmag .= "uncertainty, "; $vnetmag .= "$x, "; }
            $x = substr($s, 15, 4); $x =~ s/\ //g;
            if ($x ne "") { $qnetmag .= "nsta, "; $vnetmag .= "$x, "; }
            $t = substr($s, 20, 9); $t =~ s/\ //g;
            if ($t ne "") { $qnetmag .= "auth, "; $vnetmag .= "\'$t\', "; }
            $qnetmag .= "lddate"; $vnetmag .= "sysdate";
            $qnetmag .= ") $vnetmag)";
            sql_do($dbh, $qnetmag) || die "Failed query: $qnetmag\n";
        }
        if ($i < $n - 1) {
            $qorig .= "lddate"; $vorig .= "sysdate";
            $qorig .= ") $vorig)";
            sql_do($dbh, $qorig) || die "Failed query: $qorig\n";
        }
    }
#
#   parse and insert assocs, arrivals and amplitudes
#
    $nass = 0;
    ($oday, undef) = split ' ', $htime;
    for $s (@$phaseblock) {
#
#       skip arrivals without station or arrival time specified
#
        $sta = substr($s, 0, 5); $sta =~ s/\ //g;
        $htime = substr($s, 28, 12); $htime =~ s/\ //g;
        next if ($sta eq "" or $htime eq "");
#
#       assoc and arrival records
#
        $arid = sql_lastid($dbh, "lastid", "arid") + 1;
        $qassoc = "insert into $pars{assoc_table} (orid, arid, sta, ";
        $vassoc = "values ($orid, $arid, \'$sta\', ";
        $qarriv = "insert into $pars{arrival_table} (arid, sta, ";
        $varriv = "values ($arid, \'$sta\', ";
        $y = substr($s, 144, 5); $y =~ s/\ //g;
        $pauth = ($y eq "") ? $auth : $y;
        $qarriv .= "auth, "; $varriv .= "\'$pauth\', ";
        $delta = substr($s, 6, 6); $delta =~ s/\ //g;
        if ($delta ne "") { $qassoc .= "delta, "; $vassoc .= "$delta, "; }
        $x = substr($s, 13, 5); $x =~ s/\ //g;
        if ($x ne "") { $qassoc .= "esaz, "; $vassoc .= "$x, "; }
        $phase = substr($s, 19, 8); $phase =~ s/\ //g;
        if ($phase ne "") { $qassoc .= "phase, "; $vassoc .= "\'$phase\', "; }
        if ($phase ne "") { $qarriv .= "iphase, "; $varriv .= "\'$phase\', "; }
        $atime = human_to_epoch("$oday $htime");
        $atime += 86400. if ($atime < $etime - 3600);
        ($htime, $jdate, undef) = epoch_to_human($atime);
        $x = ($phase eq "I") ? 8 : $pars{deltim};
        $qarriv .= "time, jdate, deltim, "; $varriv .= "$atime, $jdate, $x, ";
        $x = substr($s, 41, 5); $x =~ s/\ //g;
        if ($x ne "") { $qassoc .= "timeres, "; $vassoc .= "$x, "; }
        $y = substr($s, 47, 5); $y =~ s/\ //g;
        if ($y ne "") {
            $x = ($phase eq "I" or $phase eq "H" or $phase eq "O") ? 1 : $pars{delaz};
            $qarriv .= "azimuth, delaz, "; $varriv .= "$y, $x, ";
        }
        $x = substr($s, 53, 5); $x =~ s/\ //g;
        if ($x ne "") { $qassoc .= "azres, "; $vassoc .= "$x, "; }
        $y = substr($s, 59, 5); $y =~ s/\ //g;
        if ($y ne "") {
            $x = $pars{delslo};
            $qarriv .= "slow, delslo, "; $varriv .= "$y, $x, ";
        }
        $x = substr($s, 66, 5); $x =~ s/\ //g;
        if ($x ne "") { $qassoc .= "slores, "; $vassoc .= "$x, "; }
        $isdef = substr($s, 73, 1);
        $t = ($isdef eq 'T') ? 'd' : 'n';
        $qassoc .= "timedef, "; $vassoc .= "\'$t\', ";
        $isdef = substr($s, 74, 1);
        $t = ($isdef eq 'A') ? 'd' : 'n';
        $qassoc .= "azdef, "; $vassoc .= "\'$t\', ";
        $isdef = substr($s, 75, 1);
        $t = ($isdef eq 'S') ? 'd' : 'n';
        $qassoc .= "slodef, "; $vassoc .= "\'$t\', ";
        $x = substr($s, 77, 5); $x =~ s/\ //g;
        if ($x ne "") { $qarriv .= "snr, "; $varriv .= "$x, "; }
        $t = substr($s, 101, 1); $t =~ s/\ //g;
        if ($t ne "") { $qarriv .= "qual, "; $varriv .= "\'$t\', "; }
#
#       amplitude and stamag records
#
        $amp = substr($s, 83, 9); $amp =~ s/\ //g;
        $per = substr($s, 93, 5); $per =~ s/\ //g;
        $amp += 0; $per += 0;
        if ($amp) {
            $isamp = 1;
            $ampid = sql_lastid($dbh, "lastid", "ampid") + 1;
            $qamplit = "insert into $pars{amplitude_table} (ampid, arid, amp, ";
            $vamplit = "values ($ampid, $arid, $amp, ";
            $qarriv .= "amp, "; $varriv .= "$amp, ";
            if ($per > 0) {
                $qamplit .= "per, "; $vamplit .= "$per, ";
                $x = log10($amp / $per);
                $qarriv .= "per, logat, "; $varriv .= "$per, $x, ";
            }
            $t = "A5";
            $t = "LR" if (uc($phase) eq "LR");
            $qamplit .= "amptype, inarrival, auth, ";
            $vamplit .= "\'$t\', \'y\', \'$pauth\', ";
            $qamplit .= "lddate"; $vamplit .= "sysdate";
            $qamplit .= ") $vamplit)";
            sql_do($dbh, $qamplit) || die "Failed query: $qamplit\n";
        }
        $mtyp = substr($s, 103, 5); $mtyp =~ s/\ //g;
        $y = substr($s, 109, 4); $i =~ s/\ //g;
        $y += 0;
        if ($y > 0) {
            $qstamag = "insert into $pars{stamag_table} (evid, orid, arid, ";
            $vstamag = "values ($evid, $orid, $arid, ";
            if ($isamp) { $qstamag .= "ampid, "; $vstamag .= "$ampid, "; }
            $x = $smagid;
            $x = $msid if (uc($mtyp) =~ /^MS/);
            $x = $mlid if (uc($mtyp) =~ /^ML/);
            $x = $mbid if (uc($mtyp) =~ /^MB/);
            $qstamag .= "magid, magnitude, magtype, ";
            $vstamag .= "$x, $y, \'$mtyp\', ";
            $qstamag .= "sta, delta, phase, magdef, auth, ";
            $vstamag .= "\'$sta\', $delta, \'$phase\', \'d\', \'$pauth\', ";
            $qstamag .= "lddate"; $vstamag .= "sysdate";
            $qstamag .= ") $vstamag)";
            sql_do($dbh, $qstamag) || die "Failed query: $qstamag\n";
        }
        $qassoc .= "lddate"; $vassoc .= "sysdate";
        $qassoc .= ") $vassoc)";
        sql_do($dbh, $qassoc) || die "Failed query: $qassoc\n";
        $qarriv .= "lddate"; $varriv .= "sysdate";
        $qarriv .= ") $varriv)";
        sql_do($dbh, $qarriv) || die "Failed query: $qarriv\n";
        $nass++;
    }
#
#   preferred origin
#
    $qorig .= "nass, lddate"; $vorig .= "$nass, sysdate";
    $qorig .= ") $vorig)";
    sql_do($dbh, $qorig) || die "Failed query: $qorig\n";
}

#
#
# sql_do($dbh, $query);
#
#   executes sql update/insert statement
#   returns 0 on error, 1 otherwise
#
sub sql_do {
    my ($dbh, $query) = @_;
    $dbh->do($query) or die "$query\n", $dbh->errstr;
    return 1;
}

#
#
# sql_lastid($dbh, $table, $keyname)
#    $orid = sql_lastid($dbh, $lastid, "orid") + 1;
#
sub sql_lastid {
    my ($dbh, $table, $keyname) = @_;
    my ($query, $lastid, $csr, @row);
#
#   get the lastid value
#
    $query = "select keyvalue from $table where keyname = '$keyname'
                 for update of keyvalue";
    $csr = $dbh->prepare($query) || die $dbh->errstr;
    $csr->execute || die "$query\n".$dbh->errstr;
    while (@row = $csr->fetchrow_array()) {
        $lastid = $row[0];
    }
    $csr->finish;
#
#   update the lastid table to the new value
#
    $query = "update $table set keyvalue = keyvalue + 1 where keyname = '$keyname'";
    $dbh->do($query) || die "$query\n", $dbh->errstr;
    return $lastid;
}

#
# Epoch to human time
#   ($htime,$jdate,$yr,$mon,$mday,$hr,$mn,$sec,$dsec)=epoch_to_human($etime);
#
sub epoch_to_human {
    my ($epoch) = @_;
    my ($sec, $ss, $mn, $hr, $mday, $mon, $yr, $wday, $yday, $dst, $it);
    my ($dsec, $jdate, $htime, $etime, $islastminute);
    my $sep = "-";
    $etime = sprintf "%.3f", $epoch;
    $it = int($etime);
    $dsec = sprintf "%.3f" , ($etime - int($etime));
    if ($it == 0 && $dsec < 0) {
        ($sec, $mn, $hr, $mday, $mon, $yr, $wday, $yday, $dst) = gmtime(-1);
    }
    else {
        ($sec, $mn, $hr, $mday, $mon, $yr, $wday, $yday, $dst) = gmtime($it);
    }
    $yr += 1900;
    $jdate = 1000 * $yr + $yday + 1;
    $ss = $sec + $dsec;
    $ss += 1 if ($it == 0 && $dsec < 0);
    $dsec += 1 if ($dsec < 0);
    $dsec = int(1000. * $dsec);
    $htime = sprintf "%04d%s%02d%s%02d %02d:%02d:%06.3f",
             $yr, $sep, $mon + 1, $sep, $mday, $hr, $mn, $ss;
    return $htime, $jdate, $yr, $mon + 1, $mday, $hr, $mn, $sec, $dsec;
}

#
# Human to epoch time
#    $etime = human_to_epoch($htime);
#
sub human_to_epoch {
    my $htime = shift;
    my ($date, $time, $yr, $mon, $mday, $hr, $mn, $sec, $dsec, $etime);
    ($date, $time) = split /\s+/, $htime;
    ($yr, $mon, $mday) = split /[\/\-]/, $date;
    $mon = int($mon);
    $mday = int($mday);
    ($hr, $mn, $sec) = split /:/, $time;
    $hr = int($hr);
    $mn = int($mn);
    if ($sec =~ /(\d+)(\.\d+)/) {
        $sec = int($1);
        $dsec = $2;
    } else {
        $sec = int($sec);
        $dsec = 0;
    }
    $etime = timegm($sec, $mn, $hr, $mday, $mon - 1, $yr) + $dsec;
    return $etime;
}
