#!/usr/bin/perl
#
# Convert Nordic (Seisan) format to ISF
#

$| = 1;
use strict;
use Gregion;

my $pi = 4 * atan2(1,1);
my $DEG2RAD = $pi / 180;
my $RAD2DEG = 180 / $pi;
my $RADIUS_EARTH = 6371.0;
my $DEG2KM = $DEG2RAD * $RADIUS_EARTH;

if (@ARGV < 2) {
    print "\nUsage: $0 nordicfile isffile\n";
    exit 1;
}
my ($infile, $outfil) = @ARGV;

my ($line, $oline, $mblock);
my ($evid, $orid, $arid, $k, $isok);

open(IN, "$infile") || die "Can't open $infile!\n";
open(OU, "> $outfil") || die "Can't open $outfil!\n";
print OU "BEGIN IMS1.0\n";
print OU "DATA_TYPE BULLETIN IMS1.0:short\n";
$evid = $orid = $arid = 1;
while ($line = <IN>) {
    $k = substr($line, 79, 1);
    $isok = 0;
    $mblock = "";
    if ($k == 1 or $k eq ' ') {
#
#       hypocenter line
#
        printf OU "\nEvent %s \n\n", $evid;
        print OU "   Date       Time        Err   RMS Latitude Longitude  ";
        print OU "Smaj  Smin  Az Depth   Err Ndef Nsta Gap  mdist  Mdist ";
        print OU "Qual   Author      OrigID\n";
        ($oline, $mblock) = ParseHypocenter($line, $mblock, $orid);
        print OU "$oline";
        $orid++;
        $evid++;
        while ($line = <IN>) {
            $k = substr($line, 79, 1);
            if ($k == 1) {
                ($oline, $mblock) = ParseHypocenter($line, $mblock, $orid);
                print OU "$oline";
                $orid++;
                next;
            }
            if ($k == 2 or $k == 3 or $k == 5 or $k == 6 or
                $k eq "E" or $k eq "F" or $k eq "H" or $k eq "I" or
                $k eq "M" or $k eq "P" or $line =~ /MACRO/) {
#
#               skip these lines
#
                next;
            }
            if ($k == 7) {
                if ($mblock ne "") {
                    print OU "\nMagnitude  Err Nsta Author      OrigID\n";
                    print OU "$mblock";
                }
                print OU "\nSta     Dist  EvAz Phase        Time      TRes  ";
                print OU "Azim AzRes   Slow   SRes Def   SNR       Amp   Per ";
                print OU "Qual Magnitude    ArrID\n";
#
#               phase lines (assumes that line 7 exists!)
#
                while ($line = <IN>) {
                    $isok = 1 if ($line =~ /^$/);
                    last if ($isok);      # empty line ends the phase block
                    $oline = ParsePhase($line, $arid);
                    print OU "$oline\n";
                    $isok = 1 if ($oline eq "");
                    last if ($isok);
                    $arid++;
                }
            }
            last if ($isok);      # empty line ends the phase block
        }
    }
}
print OU "\nSTOP\n";
close OU;
close IN;
print "done\n";

#
# Parse hypocenter line
#
sub ParseHypocenter {
    my ($line, $mblock, $orid) = @_;
    my ($yr, $mm, $dd, $hr, $mn, $ss, $ofix, $dfix, $lat, $lon, $depth);
    my ($typ, $auth, $nsta, $rms, $t, $mag, $oline);
#
#   origin line
#
    $yr = substr($line, 1, 4);
    $mm = substr($line, 6, 2);
    $dd = substr($line, 8, 2);
    $ofix = substr($line, 10, 1);
    $hr = substr($line, 11, 2);
    $mn = substr($line, 13, 2);
    $ss = substr($line, 16, 4);
    $t = substr($line, 22, 1);
    $lat = substr($line, 23, 7);
    $lon = substr($line, 30, 8);
    $depth = substr($line, 38, 5);
    $dfix = substr($line, 43, 1);
    $auth = substr($line, 45, 3);
    $nsta = substr($line, 48, 3);
    $rms = substr($line, 51, 4);
    $rms = "      " if ($rms < 0);
    $typ = "  ";
    $typ = "kx" if ($t eq "E");
    $typ = "sx" if ($t eq "P");
    $typ = "kv" if ($t eq "V");
    $typ = "ke" if ($t eq "Q");
    $typ = "se" if ($t eq " ");
    $typ = "ls" if ($t eq "L");
    $oline = sprintf "%04d/%02d/%02d %02d:%02d:%05.2f%s       %5.2f ",
             $yr, $mm, $dd, $hr, $mn, $ss, $ofix, $rms;
    $oline .= sprintf "%8.4f %9.4f                 %5.1f%s           %4d ",
              $lat, $lon, $depth, $dfix, $nsta;
    $oline .= sprintf "                      %-2s %-9s %8d\n",
              $typ, $auth, $orid;
#
#   magnitude block
#
    $mag = substr($line, 55, 4);
    $t = substr($line, 59, 1);
    $auth = substr($line, 60, 3);
    $typ = "";
    $typ = "ML" if ($t eq "L");
    $typ = "mb" if ($t eq "b");
    $typ = "mB" if ($t eq "B");
    $typ = "Ms" if ($t eq "s");
    $typ = "MS" if ($t eq "S");
    $typ = "MW" if ($t eq "W");
    $typ = "Mc" if ($t eq "C");
    $typ = "MbLg" if ($t eq "G");
    if ($typ ne "") {
        $mblock .= sprintf "%-5s %4.1f          %-9s %d\n",
                           $typ, $mag, $auth, $orid;
    }
    $mag = substr($line, 63, 4);
    $t = substr($line, 67, 1);
    $auth = substr($line, 68, 3);
    $typ = "";
    $typ = "ML" if ($t eq "L");
    $typ = "mb" if ($t eq "b");
    $typ = "mB" if ($t eq "B");
    $typ = "Ms" if ($t eq "s");
    $typ = "MS" if ($t eq "S");
    $typ = "MW" if ($t eq "W");
    $typ = "Mc" if ($t eq "C");
    $typ = "MbLg" if ($t eq "G");
    if ($typ ne "") {
        $mblock .= sprintf "%-5s %4.1f          %-9s %d\n",
                           $typ, $mag, $auth, $orid;
    }
    $mag = substr($line, 71, 4);
    $t = substr($line, 75, 1);
    $auth = substr($line, 76, 3);
    $typ = "";
    $typ = "ML" if ($t eq "L");
    $typ = "mb" if ($t eq "b");
    $typ = "mB" if ($t eq "B");
    $typ = "Ms" if ($t eq "s");
    $typ = "MS" if ($t eq "S");
    $typ = "MW" if ($t eq "W");
    $typ = "Mc" if ($t eq "C");
    $typ = "MbLg" if ($t eq "G");
    if ($typ ne "") {
        $mblock .= sprintf "%-5s %4.1f          %-9s %d\n",
                           $typ, $mag, $auth, $orid;
    }
    return $oline, $mblock;
}

#
# Parse phase line
#
sub ParsePhase {
    my ($line, $arid) = @_;
    my ($q, $sta, $p, $phase, $hr, $mn, $ss, $azim, $ares, $tres, $dist, $esaz);
    my $oline = "";
#
#   phase line
#
    $sta = substr($line, 1, 5);
    $sta =~ s/ //g;
    return $oline if ($sta eq "");  # end of phase block!
    $q = lc(substr($line, 9, 1));
    $p = substr($line, 10, 8);
    ($phase, undef) = split ' ', $p;
    $hr = substr($line, 18, 2);
    $hr -= 24 if ($hr > 23);
    $mn = substr($line, 20, 2);
    $ss = substr($line, 22, 6);
    $azim = substr($line, 46, 5);
    $azim =~ s/ //g;
    $ares = substr($line, 60, 3);
    $tres = substr($line, 63, 5);
    $dist = substr($line, 70, 5);
    $dist /= $DEG2KM;
    $esaz = substr($line, 76, 3);
    $oline .= sprintf "%-5s %6.2f %5.1f %-8s %02d:%02d:%06.3f ",
             $sta, $dist, $esaz, $phase, $hr, $mn, $ss;
    $tres = 99 if ($tres > 998);
    $tres = -99 if ($tres < -998);
    $oline .= sprintf "%5.1f ", $tres;
    if ($azim eq "") {
        $oline .= "                          ";
    }
    else {
        $oline .= sprintf "%5.1f %5.1f               ", $azim, $ares;
    }
    $oline .= sprintf "T__                       --%s            %8d", $q, $arid;
    return $oline;
}


