#!/usr/bin/perl
#
# Istvan Bondar, 2018
# ibondar2014@gmail.com
#
# StationListToIDCOracleDB.pl
#     Loads an ISC station list file into the IDC Oracle DB schema.
# Tables populated:
#     affiliation
#     site
# Updates the IR2_stalist file that used by iLoc for ISF2 input.
#
# StationListFile:
#    sta, altsta, lat, lon, elev
#
$|=1;
use strict;
use DBI;

my ($infile, $machine, $usr, $passwd, $dbh);
my ($sta, $prista, $lat, $lon, $elev, $s, $n, $k, $line);
my ($sta, $slat, $slon, $selev, $naff, $year, $doy);
my (@stations, @row, $query, $ondate, $offdate);

if (@ARGV < 4) {
    print "Usage: StationListToIDCOracleDB.pl StationFile machine user passwd\n";
    print "StationFile is an ISC-style station list:\n";
    print "   sta, altsta, latitude, longitude, elevation\n";
    exit -1;
}
($infile, $machine, $usr, $passwd) = @ARGV;
#
# Connect to the IDC database
#
$dbh = DBI->connect("dbi:Oracle:$machine", $usr, $passwd) or
            die "Not able to connect to database $usr\@$machine\n";
#
#
# Read stationfile
#     sta, altsta, lat, lon, elev
#
#     in the IDC DB schema elev is in km!
#
open(IN, $infile) || die "Cannot open file!\n";
open(OU, "> IR2.IR_stalist") || die "Cannot open file!\n";
while (<IN>) {
    s/\r\n$/\n/;      # replace CRLF with LF
    ($sta, $prista, $lat, $lon, $elev) = split ', ';
    chomp($elev);
    $line = "FDSN.IR.$sta.--:$lat $lon $elev";
    print OU "$line\n";
    $s .= "$line\n* <EOE>\n";
    push @stations, [ ($sta, $lat, $lon, $elev / 1000.) ];
    $n++;
}
close OU;
close IN;
print "$n IR stations\n";
#
#
# update/insert stations into IDC Oracle database schema
#
#
for ($k = 0; $k < $n; $k++) {
    ($sta, $lat, $lon, $elev) = @{$stations[$k]};
#
#   check affiliation
#
    $query = "select count(*), to_char(sysdate,'yyyyddd')
                from affiliation where sta = '$sta'";
    chomp(@row = sql_select($dbh, $query));
    for (@row) {
        ($naff, $doy) = split;
    }
#
#   insert affiliation records if necessary
#
    if ($naff) {
        $query = "select count(*) from affiliation
                   where net = 'IR' and sta = '$sta'";
        chomp(@row = sql_select($dbh, $query));
        if ($row[0] == 0) {
            $query = "insert into affiliation (net, sta, lddate)
                      values ('IR', '$sta', sysdate)";
            sql_do($dbh, $query) or die "$query failed!\n";
        }
    }
    else {
        $query = "insert into affiliation (net, sta, lddate)
                  values ('IR', '$sta', sysdate)";
        sql_do($dbh, $query) or die "$query failed!\n";
        $query = "insert into affiliation (net, sta, lddate)
                  values ('$sta', '$sta', sysdate)";
        sql_do($dbh, $query) or die "$query failed!\n";
    }
#
#   insert/update site
#
    if ($naff) {
#
#       insert new record into site if coordinates differ
#
        $query = "select lat, lon, elev, ondate, offdate
                    from site where sta = '$sta' order by ondate";
        chomp(@row = sql_select($dbh, $query));
        for (@row) {
            ($slat, $slon, $selev, $ondate, $offdate) = split;
        }
        if (abs($lat - $slat) > 0.0001 ||
            abs($lon - $slon) > 0.0001 ||
            abs($elev - $selev) > 0.0001) {
            $query = "insert into site
                             (sta, ondate, offdate, lat, lon, elev, staname,
                              statype, refsta, dnorth, deast, lddate)
                     (select sta, $doy, -1, $lat, $lon, $elev, staname,
                             statype, refsta, dnorth, deast, sysdate
                        from site where sta = '$sta'
                         and ondate = $ondate
                         and offdate = $offdate)";
            sql_do($dbh, $query) or die "$query failed!\n";
            $query = "update site set offdate = $doy
                       where sta = '$sta'
                         and ondate = $ondate
                         and offdate = $offdate";
            sql_do($dbh, $query) or die "$query failed!\n";
        }
    }
    else {
#
#       insert new site record
#
        $query = "insert into site
                         (sta, ondate, offdate, lat, lon, elev, statype, lddate)
                  values ('$sta', to_date(1970-01-01', 'yyyy-mm-dd'), -1,
                          $lat, $lon, $elev, 'ss', sysdate)";
        sql_do($dbh, $query) or die "$query failed!\n";
    }
}
print "updated IDC Oracle database\n";
#
# Disconnect from the database
#
$dbh->disconnect;
#
#
# update IR section in IR2_stalist
#
#
if (-e IR2_stalist) {
    `cp IR2_stalist IR2_stalist.old`;
    $k = 1;
    open(IN, "IR2_stalist.old") || die "Cannot open file!\n";
    open(OU, "> IR2_stalist") || die "Cannot open file!\n";
    while ($line = <IN>) {
        if ($line =~ /FDSN\.IR\./) {
            print OU $s if ($k);
            <IN>;
            $k = 0;
        }
        else {
            print OU $line
        }
    }
    close IN;
    close OU;
}
else {
    open(OU, "> IR2_stalist") || die "Cannot open file!\n";
    print OU $s;
    close OU;
}
print "updated IR section in IR2_stalist\n";
print "done\n";

#
#  sql_select($dbh, $query)
#
#   executes sql select statement
#   returns each row (with newline) as a separate element of @rows
#   (list of strings)
#
sub sql_select {
    my ($dbh, $query) = @_;
    my ($csr, @row, $line);
    my @rows = ();
    $csr = $dbh->prepare($query) || die $dbh->errstr;
    $csr->execute || die "$query\n".$dbh->errstr;
    while (@row = $csr->fetchrow_array) {
        push @rows, "@row\n";
    }
    $csr->finish;
    return @rows;
}

#
#  sql_do($dbh, $query)
#
#   executes sql update/insert statement
#   returns 0 on error, 1 otherwise
#
#
sub sql_do {
    my ($dbh, $query) = @_;
    $dbh->do($query) or die "$query\n", $dbh->errstr;
    return 1;
}
